import * as THREE from '../build/three.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { EffectComposer } from './jsm/postprocessing/EffectComposer.js';
import { RenderPass } from './jsm/postprocessing/RenderPass.js';
import { ShaderPass } from './jsm/postprocessing/ShaderPass.js';
import { LuminosityShader } from './jsm/shaders/LuminosityShader.js';
import { SobelOperatorShader } from './jsm/shaders/SobelOperatorShader.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let scene = new THREE.Scene();

        let renderer = new THREE.WebGLRenderer({ antialias: true });
        this.domWebGL.appendChild(renderer.domElement);

        this.renderer = renderer;
        this.scene = scene;

        this._setupCamera();
        this._setupModel();
        this._setupLights();
        this._setupComposers();
        this._setupControls();
        this._setupGUI();

        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupComposers() {
        const composer = new EffectComposer(this.renderer);
        
        const renderPass = new RenderPass(this.scene, this._camera);
        composer.addPass(renderPass);

        const effectGrayScale = new ShaderPass(LuminosityShader);
        composer.addPass(effectGrayScale);

        const effectSobel = new ShaderPass(SobelOperatorShader);
        effectSobel.uniforms['resolution'].value.x = window.innerWidth * window.devicePixelRatio;
        effectSobel.uniforms['resolution'].value.y = window.innerHeight * window.devicePixelRatio;
        composer.addPass(effectSobel);

        this._composer = composer;
        this._effectSobel = effectSobel;
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this.renderer.domElement);
        controls.minDistance = 10;
        controls.maxDistance = 100;
    }

    _setupModel() {
        const geometry = new THREE.TorusKnotGeometry(8, 3, 256, 32, 4, 6);
        const material = new THREE.MeshPhongMaterial({ color: 0xffff00 });
        const mesh = new THREE.Mesh(geometry, material);

        this.scene.add(mesh);
    }

    _setupLights() {
        const ambientLight = new THREE.AmbientLight(0xcccccc, 0.4);
        this.scene.add(ambientLight);

        const pointLight = new THREE.PointLight(0xffffff, 0.8);
        this._camera.add(pointLight);
        this.scene.add(this._camera);
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(70, window.innerWidth / window.innerHeight, 0.1, 200);
        camera.position.set(0, 10, 25);
        camera.lookAt(this.scene.position);
        this.scene.add(camera);

        this._camera = camera;
    }

    update() {
        //.
    }

    render(time) {
        requestAnimationFrame(this.render.bind(this));
        this.update(time);

        if(this._params.enable) {
            this._composer.render();
        } else {
            this.renderer.render(this.scene, this._camera);
        }
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setSize(window.innerWidth, window.innerHeight);

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._composer.setSize(window.innerWidth, window.innerHeight);
        this._effectSobel.uniforms['resolution'].value.x = window.innerWidth * window.devicePixelRatio;
        this._effectSobel.uniforms['resolution'].value.y = window.innerHeight * window.devicePixelRatio;
        
        this.render();
    }

    _setupGUI() {
        const params = {
            'enable': true
        };

        const gui = new GUI();
        gui.add(params, 'enable');

        gui.open();

        this._params = params;
    }
}

window.onload = function () {
    (new App()).render();
}